unit MyGSIF;

interface

	uses
		Types, Files, QDOffscreen;
		
	function WriteImageToGSIF( source: GWorldPtr; var data: Handle ): OSErr;
	procedure DrawGSIFWithin( data: Handle; dest: PixMapHandle; ath, atv: integer; const bounds: Rect );
	procedure DrawGSIF( data: Handle; dest: PixMapHandle; ath, atv: integer );
	procedure GetGSIFSize( gsifdata: Handle; var width, height: integer );
	
implementation

	uses
		MyAssertions, MyMathUtils, MyLowLevel, MyMemory, MyUtils;

{$setc display_methods := false}
{$setc Pascal_DoLines := display_methods or false}

	type
		GSIFHeader = packed record
			tag: OSType;
			width: integer;
			height: integer;
		end;
	
	const
		GSIF_Header_tag = 'GSIF';

(*
Format:
	2b - width
	2b - height
	height list of rows
row: 
	2b - data length
	list of blocks
block:
	either:
	-	1b (0-63?) + bytes - raw data, bytes - Raw
	-	1b (128-250) + byte - RLE
	-	1b (251-255) + 1b length + byte - RLE2
	-	1b (64-95) + 2b + bitbytes - two colour map - 2
	-	1b (96-126) + 4b + bitbytes - four colour map - 4
	-	1b (127) + 1b length + 4b + bitbytes - four colour map - 42
*)

	type
		methods= (M_Raw, M_RLE, M_RLE2, M_2, M_4, M_42 );
	const
		M_First = M_Raw;
		M_Last = M_42;
	const
		min_raw_count = 1;
		max_raw_count = 63 - 0 + min_raw_count;
		min_RLE1_count = 3;
		max_RLE1_count = 250 - 128 + min_RLE1_count;
		min_RLE2_count = max_RLE1_count + 1;
		max_RLE2_count = min_RLE2_count + 5 * 256 - 1;
		min_2_count = 8;
		max_2_count = min_2_count + 8 * (95-64);
		min_4_count = 12;
		max_4_count = min_4_count + 4 * (126-96);
		min_42_count = max_4_count + 4;
		max_42_count = min_42_count + 4 * 255;

	procedure AssertPtrInGWorld( gworld: GWorldPtr; p: Ptr );
		var
			pm: PixMapHandle;
			offset: longint;
	begin
		Assert( gworld <> nil );
		pm := GetPortPixMap( gworld );
		Assert( pm <> nil );
		offset := longint(p) - longint( GetPixBaseAddr( pm ) );
		Assert( offset >= 0 );
		Assert( offset <= RectHeight( gworld^.portRect ) * band( pm^^.rowBytes, $7FFF ) );
		Assert( offset mod band( pm^^.rowBytes, $7FFF ) <= RectWidth( gworld^.portRect ) );
	end;
	
	function WriteImageToGSIF( source: GWorldPtr; var data: Handle ): OSErr;
		var
			total_used: longint;
			total_atem: array[methods] of longint;
			total_usedm: array[methods] of longint;
			cur_pos: longint;
	
		var
			raw_outstanding: longint;

		procedure AppendByte( b: integer );
		begin
			Assert( cur_pos < GetHandleSize( data ) );
			SetUnsignedByte( data^, cur_pos, b );
			Inc(cur_pos);
		end;
		
		procedure AppendWord( n: integer );
		begin
			AppendByte( n shr 8 );
			AppendByte( n );
		end;
		
		procedure SetWordAt( pos: longint; n: integer );
		begin
			Assert( (0 <= pos) & (pos <= GetHandleSize(data) - 2) );
			BlockMoveData( @n, AddPtrLong( data^, pos ), 2 );
		end;
		
		procedure FlushRaw( addr: Ptr );
			var
				count, i: longint;
				p: Ptr;
		begin
			p := AddPtrLong( addr, -raw_outstanding );
			while raw_outstanding > 0 do begin
				count := Min( raw_outstanding, max_raw_count );
				AppendByte( count - min_raw_count );
				for i := 0 to count - 1 do begin
					AppendByte( GetUnsignedByte( p, i ) );
				end;
				OffsetPtr( p, count );
				raw_outstanding := raw_outstanding - count;

				total_used := total_used + (1+count);
				total_usedm[M_Raw] := total_usedm[M_Raw] + (1+count);
				total_atem[M_Raw] := total_atem[M_Raw] + count;
			end;
		end;
		
		procedure Try( addr: Ptr; len: longint; var did: longint );
				
			function CountRuns( count: longint ): longint;
				var
					b, last: integer;
					runs, i: longint;
			begin
				last := -1;
				runs := 0;
				for i := 0 to count - 1 do begin
					b := GetUnsignedByte( addr, i );
					if b <> last then begin
						Inc(runs);
						last := b;
					end;
				end;
				CountRuns := runs;
			end;
			
			function CheckColours( pos: longint; colours: integer; min: longint ): boolean;
				var
					i: longint;
					b, b1, b2, b3, b4: integer;
			begin
				Assert( (0 < colours) & (colours <= 4) );
				CheckColours := true;
				b1 := GetUnsignedByte( addr, pos );
				b2 := b1;
				b3 := b1;
				b4 := b1;
				for i := 1 to min - 1 do begin
					b := GetUnsignedByte( addr, pos + i );
					if (b <> b1) & (b <> b2) & (b <> b3) & (b <> b4) then begin
						if (colours > 1) & (b2 = b1) then begin
							b2 := b;
						end else if (colours > 2) & (b3 = b1) then begin
							b3 := b;
						end else if (colours > 3) & (b4 = b1) then begin
							b4 := b;
						end else begin
							CheckColours := false;
							leave;
						end;
					end;
				end;
			end;
			
			var
				bestused, bestate: longint;
				used, ate: longint;
				b, bout, b1, b2, b3, b4: integer;
				bb1, bb2, bb3, bb4: integer;
				count: longint;
				i, j: longint;
				method: methods;
		begin
			Assert( len > 0 );
			// raw bytes;
			bestate := Min( len, 64 );
			bestused := bestate + 1;
			method := M_Raw;
			
			if len >= min_RLE1_count then begin
				b := GetUnsignedByte( addr, 0 );
				count := 1;
				while (count <= len) & (GetUnsignedByte( addr, count ) = b) do begin
					Inc(count);
				end;
				if count > len then begin
					count := len;
				end;
				if count >= min_RLE1_count then begin
					if count > max_RLE1_count then begin
						count := Min( count, max_RLE2_count );
						bestused := 3;
						bestate := count;
						method := M_RLE2;
					end else begin
						bestate := count;
						bestused := 2;
						method := M_RLE;
					end;
				end;
			end;
			
			if (len >= min_2_count) & ((method = M_Raw) | (bestate < min_2_count)) & not CheckColours( 0, 1, 16 ) then begin
				b1 := GetUnsignedByte( addr, 0 );
				b2 := b1;
				for i := 1 to min_2_count do begin { must be in there or we'd have picked it up as a RLE }
					if b2 = b1 then begin
						b2 := GetUnsignedByte( addr, i )
					end;
				end;
				Assert( b1 <> b2 );
				count := 0;
				while (count <= len) & ((GetUnsignedByte( addr, count ) = b1) | (GetUnsignedByte( addr, count ) = b2)) do begin
					Inc(count);
				end;
				if count > len then begin
					count := len;
				end;
				count := count div 8 * 8;
				
				if count >= 16 then begin
					if CheckColours( count - 16, 1, 16 ) then begin
						b := GetUnsignedByte( addr, count - 1 );
						while (count >= 8) & (GetUnsignedByte( addr, count - 1 ) = b) & CheckColours( count - 8, 1, 8 ) do begin
							count := count - 8;
						end;
					end;
				end;
				
				if count > 40 then begin // Check for long run
					for i := 1 to count - 40 do begin
						if (i mod 8 = 0) & CheckColours( i, 1, 40 ) then begin
							count := i;
							leave;
						end;
					end;
				end;
				
				if count >= min_2_count then begin
					if count > max_2_count then begin
						count := max_2_count;
					end;
					if 2 * CountRuns( count ) > 3 + count div 8 then begin
						bestate := count;
						bestused := 3 + count div 8;
						bb1 := b1;
						bb2 := b2;
						method := M_2;
					end;
				end;
			end;
			
			if (len >= 12) & not CheckColours( 0, 1, 8 ) then begin
				b1 := GetUnsignedByte( addr, 0 );
				b2 := b1;
				b3 := b1;
				b4 := b1;
				count := 0;
				while count <= len do begin
					b := GetUnsignedByte( addr, count );
					if (b <> b1) & (b <> b2) & (b <> b3) & (b <> b4) then begin
						if b2 = b1 then begin
							b2 := b;
						end else if b3 = b1 then begin
							b3 := b;
						end else if b4 = b1 then begin
							b4 := b;
						end else begin
							leave;
						end;
					end;
					Inc(count);
				end;
				if count > len then begin
					count := len;
				end;
				count := count div 4 * 4;

				if count >= 12 then begin
					if CheckColours( count - 8, 1, 8 ) then begin
						b := GetUnsignedByte( addr, count - 1 );
						while (count >= 4) & (GetUnsignedByte( addr, count - 1 ) = b) & CheckColours( count - 4, 1, 4 ) do begin
							count := count - 4;
						end;
					end;
				end;

				if count > 28 then begin // Check for long run
					for i := 1 to count - 28 do begin
						if (i mod 4 = 0) & CheckColours( i, 1, 28 ) then begin
							count := i;
							leave;
						end;
					end;
				end;
				
				if count >= min_4_count then begin
					if count > max_42_count then begin
						count := max_42_count;
					end;
					used := 5 + count div 4;
					if count > max_4_count then begin
						used := used + 1;
					end;
					ate := count;
					if 2 * CountRuns( count ) > used then begin
						if 100 * used div ate < 100 * bestused div bestate then begin
							bestate := ate;
							bestused := used;
							if count > max_4_count then begin
								method := M_42;
							end else begin
								method := M_4;
							end;
							bb1 := b1;
							bb2 := b2;
							bb3 := b3;
							bb4 := b4;
						end;
					end;
				end;
			end;
			
(*
			if method = M_Raw then begin
				for i := 1 to Min( bestate - 1, len - min_RLE1_count ) do begin
					if CheckColours( i, 1, min_RLE1_count ) then begin
						bestate := i;
						bestused := bestate + 1;
						method := M_Raw;
						leave;
					end;
				end;

				for i := 1 to Min( bestate - 1, len - min_2_count ) do begin
					if CheckColours( i, 2, min_2_count ) then begin
						bestate := i;
						bestused := bestate + 1;
						method := M_Raw;
						leave;
					end;
				end;

				for i := 1 to Min( bestate - 1, len - min_4_count ) do begin
					if CheckColours( i, 4, min_4_count ) then begin
						bestate := i;
						bestused := bestate + 1;
						method := M_Raw;
						leave;
					end;
				end;
			end;
*)

			if method <> M_Raw then begin
				FlushRaw( addr );
				did := bestate;
				total_used := total_used + bestused;
				total_usedm[method] := total_usedm[method] + bestused;
				total_atem[method] := total_atem[method] + bestate;
				case method of
					M_RLE: begin
						Assert( (min_RLE1_count <= bestate) & (bestate <= max_RLE1_count) );
						AppendByte( 128 + (bestate - min_RLE1_count) );
						AppendByte( GetUnsignedByte( addr, 0 ) );
					end;
					M_RLE2: begin
						Assert( (min_RLE2_count <= bestate) & (bestate <= max_RLE2_count) );
						AppendByte( 251 + (bestate - min_RLE2_count) div 256 );
						AppendByte( (bestate - min_RLE2_count) mod 256 );
						AppendByte( GetUnsignedByte( addr, 0 ) );
					end;
					M_2: begin
						Assert( (min_2_count <= bestate) & (bestate <= max_2_count) );
						Assert( (0 <= bb1) & (bb1 <= 255) );
						Assert( (0 <= bb2) & (bb2 <= 255) );
						Assert( bb1 <> bb2 );
						AppendByte( 64 + (bestate - min_2_count) div 8 );
						AppendByte( bb1 );
						AppendByte( bb2 );
						for i := 0 to bestate div 8 - 1 do begin
							bout := 0;
							for j := 0 to 7 do begin
								b := GetUnsignedByte( addr, i * 8 + j );
								Assert( (b = bb1) | (b = bb2) );
								bout := bout * 2;
								if b = bb2 then begin
									bout := bout + 1;
								end;
							end;
							AppendByte( bout );
						end;
					end;
					M_4, M_42: begin
						Assert( (0 <= bb1) & (bb1 <= 255) );
						Assert( (0 <= bb2) & (bb2 <= 255) );
						Assert( (0 <= bb3) & (bb3 <= 255) );
						Assert( (0 <= bb4) & (bb4 <= 255) );
						Assert( bb1 <> bb2 );
						
						if method = M_4 then begin
							Assert( (min_4_count <= bestate) & (bestate <= max_4_count) );
							AppendByte( 96 + (bestate - min_4_count) div 4 );
						end else begin
							Assert( (min_42_count <= bestate) & (bestate <= max_42_count) );
							AppendByte( 127 );
							AppendByte( (bestate - min_42_count) div 4 );
						end;
						AppendByte( bb1 );
						AppendByte( bb2 );
						AppendByte( bb3 );
						AppendByte( bb4 );
						for i := 0 to bestate div 4 - 1 do begin
							bout := 0;
							for j := 0 to 3 do begin
								b := GetUnsignedByte( addr, i * 4 + j );
								Assert( (b = bb1) | (b = bb2) | (b = bb3) | (b = bb4) );
								bout := bout * 4;
								if b = bb2 then begin
									bout := bout + 1;
								end else if b = bb3 then begin
									bout := bout + 2;
								end else if b = bb4 then begin
									bout := bout + 3;
								end;
							end;
							AppendByte( bout );
						end;
					end;
					otherwise begin
						Assert( false );
					end;
				end;
			end else begin
				Inc(raw_outstanding);
				did := 1;
			end;
		end;
		var
			source_rowBytes: longint;
			source_addr: Ptr;
			source_pm: PixMapHandle;
			row: integer;
			width, height, left: longint;
			did: longint;
			m: methods;
			dummy: boolean;
			saved_pos: longint;
			err: OSErr;
	begin
		total_used := 0;
		MZero( @total_usedm, SizeOf(total_usedm) );
		MZero( @total_atem, SizeOf(total_atem) );
		
		source_pm := GetPortPixMap( source );
		dummy := LockPixels( source_pm );
		source_addr := source_pm^^.baseAddr;
		source_rowBytes := source_pm^^.rowBytes and $3FFF;
		width := source_pm^^.bounds.right - source_pm^^.bounds.left;
		height := source_pm^^.bounds.bottom - source_pm^^.bounds.top;
		
		cur_pos := 0;
		err := MNewHandle( data, width * height * 130 div 128 + 100 );
		if err = noErr then begin
			OSTypePtr(data^)^ := GSIF_Header_tag;
			cur_pos := cur_pos + 4;
			AppendWord( width );
			AppendWord( height );
			
			
			for row := source_pm^^.bounds.top to source_pm^^.bounds.bottom - 1 do begin
				AssertPtrInGWorld( source, source_addr );
				saved_pos := cur_pos;
				AppendWord( 0 );
				
				raw_outstanding := 0;
				left := width;
				while left > 0 do begin
					Try( AddPtrLong( source_addr, width - left ), left, did );
					left := left - did;
				end;
				Assert( left = 0 );
				FlushRaw( AddPtrLong( source_addr, width ) );
				SetWordAt( saved_pos, cur_pos - saved_pos - 2 );
				OffsetPtr( source_addr, source_rowBytes );
			end;
			for m := M_First to M_Last do begin
//				DebugStr(NumToStr( ord(m) ) + ' ' + NumToStr( total_usedm[m] ) + ' ' + NumToStr( total_atem[m] ) + ';g' );
			end;
//			DebugStr( 'Total Used was ' + NumToStr( total_used ) + ';g' );
//			DebugStr( 'cur_pos was ' + NumToStr( cur_pos ) + ';g' );
			MShrinkHandleSize( data, cur_pos );
		end;
		WriteImageToGSIF := err;
	end;

{$definec Assertx(x)}
{$definec AssertxPtrInGWorld(a,b)}

(*
{$ifc display_methods}
{$definec DisplaySetUnsignedByte( a, b, c ) SetUnsignedByte( a, b, c )}
{$elsec}
{$definec DisplaySetUnsignedByte( a, b, c )}
{$endc}

	procedure WriteHandleToGSIFOld( data: Handle; dest: GWorldPtr );
		var
			cur_pos: longint;
		function GetByte: integer;
		begin
			Assertx( (0 <= cur_pos) & (cur_pos < GetHandleSize(data)) );
			GetByte := GetUnsignedByte( data^, cur_pos );
			Inc(cur_pos);
		end;
		
		procedure CheckByte( b: integer );
			var
				tb: integer;
		begin
			Assertx( (0 <= b) & (b <= 255) );
			tb := GetByte;
			Assertx( tb = b );
		end;
		
		function GetWord: integer;
			var
				b1, b2: integer;
		begin
			b1 := GetByte;
			b2 := GetByte;
			GetWord := b1 * 256 + b2; 
		end;
		
		procedure Try( addr: Ptr; var did: longint );
			var
				code, b, b1, b2, b3, b4: integer;
				count: longint;
				i, j: longint;
		begin
			code := GetByte;
			if code < 64 then begin // Raw
				count := code + 1;
				Assertx( (0 <= cur_pos) & (cur_pos + count <= GetHandleSize(data)) );
				for i := 0 to count - 1 do begin
					b := GetByte;
					SetUnsignedByte( addr, i, b ); // MACRO!
					DisplaySetUnsignedByte( addr, i, 35 );
				end;
				did := count;
			end else if code < 96 then begin // 2
				count := min_2_count + (code - 64) * 8;
				b1 := GetByte;
				b2 := GetByte;
				for i := 0 to count div 8 - 1 do begin
					b := GetByte;
					for j := 0 to 7 do begin
						if b and $80 = $00 then begin
							SetUnsignedByte( addr, i * 8 + j, b1 );
						end else begin
							SetUnsignedByte( addr,  i * 8 + j, b2 );
						end;
						DisplaySetUnsignedByte( addr, i * 8 + j, 5 );
						b := b shl 1;
					end;
				end;
				did := count;
			end else if code < 128 then begin // 4 & 42
				if code = 127 then begin
					count := min_42_count + GetByte * 4;
				end else begin
					count := min_4_count + (code - 96) * 4;
				end;
				b1 := GetByte;
				b2 := GetByte;
				b3 := GetByte;
				b4 := GetByte;
				for i := 0 to count div 4 - 1 do begin
					b := GetByte;
					for j := 0 to 3 do begin
						case b and $0C0 of
							$000: begin
								SetUnsignedByte( addr,  i * 4 + j, b1 );
							end;
							$040: begin
								SetUnsignedByte( addr,  i * 4 + j, b2 );
							end;
							$080: begin
								SetUnsignedByte( addr,  i * 4 + j, b3 );
							end;
							$0C0: begin
								SetUnsignedByte( addr,  i * 4 + j, b4 );
							end;
						end;
						DisplaySetUnsignedByte( addr, i * 4 + j, 15 );
						b := b shl 2;
					end;
				end;
				did := count;
			end else if code < 251 then begin // RLE
				count := min_RLE1_count + code - 128;
				b := GetByte;
				for i := 0 to count - 1 do begin
					SetUnsignedByte( addr, i, b );
					DisplaySetUnsignedByte( addr, i, 251 );
				end;
				did := count;
			end else begin // RLE2
				count := min_RLE2_count + 256 * (code - 251) + GetByte;
				b := GetByte;
				for i := 0 to count - 1 do begin
					SetUnsignedByte( addr, i, b );
					DisplaySetUnsignedByte( addr, i, 255 );
				end;
				did := count;
			end else begin
				Assertx( false );
			end;
		end;

		var
			width, height: longint;
			dest_rowBytes: longint;
			dest_addr: Ptr;
			dest_pm: PixMapHandle;
			dummy: boolean;
			junk, left, row, did: longint;
	begin
		cur_pos := 0;
		CheckByte( ord('P' ) );
		CheckByte( ord('N' ) );
		CheckByte( ord('L' ) );
		CheckByte( ord('i' ) );
		width := GetWord;
		height := GetWord;

		dest_pm := GetPortPixMap( dest );
		dummy := LockPixels( dest_pm );
		dest_addr := dest_pm^^.baseAddr;
		dest_rowBytes := dest_pm^^.rowBytes and $3FFF;
		Assertx( width <= dest_pm^^.bounds.right - dest_pm^^.bounds.left );
		Assertx( height <= dest_pm^^.bounds.bottom - dest_pm^^.bounds.top );
		
		for row := 0 to height - 1 do begin
			AssertxPtrInGWorld( dest, dest_addr );
			junk := GetWord;
			left := width;
			while left > 0 do begin
				Try( AddPtrLong( dest_addr, width - left ), did );
				left := left - did;
			end;
			Assertx( left = 0 );
			OffsetPtr( dest_addr, dest_rowBytes );
		end;
	end;
*)

{$ifc Pascal_DoLines}
	procedure DoLines( cur_pos:UInt8Ptr; addr: UInt8Ptr; dest_rowBytes: longint; orgwidth, skip, dowidth, height: longint );
		label
			next_line;
		var
			code, b, bt, b1, b2, b3, b4: integer;
			count: longint;
			i, j: longint;
			finaddr: UInt8Ptr;
			clipping: boolean;
			data_len: integer;
			cskip, cwidth, ccount: longint;
	begin
		Assert( (0 <= skip) & (0 <= dowidth) & (skip + dowidth <= orgwidth) );
		clipping := dowidth < orgwidth;
		while height > 0 do begin
			Dec(height);
//			AssertxPtrInGWorld( dest, dest_addr );
			BlockMoveData( cur_pos, @data_len, 2 );
			longint(cur_pos) := longint(cur_pos) + 2;
			finaddr := UInt8Ptr(longint(cur_pos) + data_len);
			cskip := skip;
			cwidth := dowidth;
			while cwidth > 0 do begin

				code := cur_pos^;
				Inc(longint(cur_pos));
				
				if code < 64 then begin // Raw
					count := code + 1;
					Assertx( (longint(data^) <= longint(cur_pos)) & (longint(cur_pos) + count <= longint(data^) + GetHandleSize(data)) );

					if cskip > 0 then begin
						ccount := Min( cskip, count );
						longint(cur_pos) := longint(cur_pos) + ccount;
						cskip := cskip - ccount;
						OffsetPtr( addr, ccount );
						count := count - ccount;
					end;

					if cskip = 0 then begin
						count := Min( count, cwidth );
						BlockMoveData( cur_pos, addr, count );
{$ifc display_methods}
						MFill( addr, count, 35 );
{$endc}
						cwidth := cwidth - count;
						longint(cur_pos) := longint(cur_pos) + count;
						OffsetPtr( addr, count );
					end;

				end else if code < 96 then begin // 2
					count := (min_2_count div 8) + (code - 64);
					
					b1 := cur_pos^;
					Inc(longint(cur_pos));
					b2 := cur_pos^;
					Inc(longint(cur_pos));
					
					if not clipping then begin
						cwidth := cwidth - count * 8;
					end else if cskip >= 8 then begin
						ccount := Min( count * 8, cskip - cskip mod 8 );
						longint(addr) := longint(addr) + ccount;
						cskip := cskip - ccount;
						ccount := ccount div 8;
						longint(cur_pos) := longint(cur_pos) + ccount;
						count := count - ccount;
					end;

					for i := 0 to count - 1 do begin
						b := cur_pos^;
						Inc(longint(cur_pos));

						for j := 0 to 7 do begin
							if b and $80 = $00 then begin
								bt := b1;
							end else begin
								bt := b2;
							end;
							if clipping then begin
								if cskip > 0 then begin
									Dec(cskip);
									Inc(longint(addr));
								end else begin
									addr^ := bt;
{$ifc display_methods}
									addr^ := 5;
{$endc}
									Inc(longint(addr));
									Dec(cwidth);
									if cwidth = 0 then begin
										goto next_line;
									end;
								end;
							end else begin
								addr^ := bt;
{$ifc display_methods}
								addr^ := 5;
{$endc}
								Inc(longint(addr));
							end;
							b := b shl 1;
						end;

					end;
				end else if code < 128 then begin // 4 & 42
					if code = 127 then begin
						count := (min_42_count div 4) + cur_pos^;
						Inc(longint(cur_pos));
					end else begin
						count := (min_4_count div 4) + (code - 96);
					end;
					b1 := cur_pos^;
					Inc(longint(cur_pos));
					b2 := cur_pos^;
					Inc(longint(cur_pos));
					b3 := cur_pos^;
					Inc(longint(cur_pos));
					b4 := cur_pos^;
					Inc(longint(cur_pos));

					if not clipping then begin
						cwidth := cwidth - count * 4;
					end else if cskip >= 4 then begin
						ccount := Min( count * 4, cskip - cskip mod 4 );
						longint(addr) := longint(addr) + ccount;
						cskip := cskip - ccount;
						ccount := ccount div 4;
						longint(cur_pos) := longint(cur_pos) + ccount;
						count := count - ccount;
					end;

					for i := 0 to count - 1 do begin
						b := cur_pos^;
						Inc(longint(cur_pos));
						for j := 0 to 3 do begin

							if b < $080 then begin
								if b < $040 then begin
									bt := b1;
								end else begin
									bt := b2;
								end;
							end else if b < $0C0 then begin
								bt := b3;
							end else begin
								bt := b4;
							end;
							
							if clipping then begin
								if cskip > 0 then begin
									Dec(cskip);
									Inc(longint(addr));
								end else begin
									addr^ := bt;
{$ifc display_methods}
									addr^ := 15;
{$endc}
									Inc(longint(addr));
									Dec(cwidth);
									if cwidth = 0 then begin
										goto next_line;
									end;
								end;
							end else begin
								addr^ := bt;
{$ifc display_methods}
								addr^ := 15;
{$endc}
								Inc(longint(addr));
							end;
							b := band(b shl 2, $0FF);
						end;
					end;
				end else begin // RLE
					if code < 251 then begin
						count := min_RLE1_count + code - 128;
{$ifc display_methods}
						bt := 251;
{$endc}
					end else begin
						count := min_RLE2_count + (code - 251) shl 8 + cur_pos^;
						Inc(longint(cur_pos));
{$ifc display_methods}
						bt := 255;
{$endc}
					end;
					b := cur_pos^;
					Inc(longint(cur_pos));

					if cskip > 0 then begin
						ccount := Min( cskip, count );
						longint(addr) := longint(addr) + ccount;
						cskip := cskip - ccount;
						count := count - ccount;
					end;

					if cskip = 0 then begin
						count := Min( count, cwidth );
						cwidth := cwidth - count;
						while count > 0 do begin
							addr^ := b;
{$ifc display_methods}
							addr^ := bt;
{$endc}
							Inc(longint(addr));
							Dec(count);
						end;
					end;
				end;
			end;
next_line:
			Assert( cwidth = 0 );
			cur_pos := finaddr;
			OffsetPtr( addr, dest_rowBytes - skip - dowidth );
		end;
	end;
	
{$elsec}
{$ifc GENERATINGPOWERPC}	
	procedure DoLines( cur_pos: UInt8Ptr; addr: UInt8Ptr; dest_rowBytes: longint; orgwidth, skip, dowidth, height: longint );
	asm;
(*
	r3 - cur_pos
	r4 - addr
	r5 - dest_rowBytes
	r6 - orgwidth
	r7 - skip
	r8 - dowidth
	r9 - height
	r10 - cskip
	r11 - cwidth
	cr7 - clipping
	r12 - b
	r13 - jcount
	r14 - bt
	r15 - b2b1 or b4b3b2b1
	r16 - count
*)
		const
			stupid_compiler_64_min_2_count_div_8 = 64-min_2_count div 8;
			stupid_compiler_96_min_4_count_div_4 = 96-min_4_count div 4;
			stupid_compiler_min_42_count_div_4 = min_42_count div 4;
			stupid_compiler_128_min_RLE1_count = 128 - min_RLE1_count;
	begin
		stw			r13,-4(sp)
		stw			r14,-8(sp)
		stw			r15,-12(sp)
		stw			r16,-16(sp)
		stw			r17,-20(sp)
		stw			r18,-24(sp)
		
		subi			r3,r3,1
		subi		r4,r4,1
		
// 		clipping := dowidth < orgwidth;
		cmpw		cr7,r8,r6
		
//		while height > 0 do begin
row_loop:

//			BlockMoveData( cur_pos, @data_len, 2 );
//			longint(cur_pos) := longint(cur_pos) + 2;
//			finaddr := UInt8Ptr(longint(cur_pos) + data_len);
		
		lbzu			r18,1(r3)
		lbzu			r17,1(r3)
		rlwimi		r17,r18,8,16,23

		add			r17,r17,r3

//			cskip := skip;
		mr			r10,r7
		
//			cwidth := dowidth;
		mr			r11,r8
		
//			while cwidth > 0 do begin
cwidth_loop:

//				code := cur_pos^;
//				Inc(longint(cur_pos));
		lbzu			r16,1(r3)

// case code of
		cmpwi		r16,64
		blt			do_raw
		cmpwi		r16,96
		blt			do_M_2
		cmpwi		r16,128
		blt			do_M_4
		b				do_RLE

do_raw:
		addi			r16,r16,1
		
		cmpwi			r10,0
		beq			no_raw_skip
		
		mr			r18,r16
		cmpw		r18,r10
		blt			min1
		mr			r18,r10
min1:
		add			r3,r3,r18
		add			r4,r4,r18
		sub			r16,r16,r18
		sub			r10,r10,r18
		cmpwi		r10,0
		bne			cwidth_fin

no_raw_skip:
		cmpw		r16,r11
		blt			min2
		mr			r16,r11
min2:
		cmpwi		r16,0
		beq			cwidth_fin
		sub			r11,r11,r16
		mtctr		r16
block_move_loop:
		lbzu			r18,1(r3)
		stbu			r18,1(r4)
		bdnz			block_move_loop
		
		b				cwidth_fin	

do_M_2:
		subi			r16,r16,63 { stupid_compiler_64_min_2_count_div_8 }

		lbzu			r15,1(r3)
		lbzu			r18,1(r3)
		rlwimi		r15,r18,8,16,23
		
		blt			cr7,are_clipping2

		rlwinm			r18,r16,3,0,28
		sub			r11,r11,r18
		b				start_2_loop

are_clipping2:
		cmpwi		r10,8
		blt			start_2_loop
		
		rlwinm		r18,r16,3,0,28
		rlwinm		r13,r10,0,0,28

		cmpw		r18,r13
		blt			min3
		mr			r18,r13
min3:
		add			r4,r4,r18
		sub			r10,r10,r18
		rlwinm		r18,r18,29,0,31
		add			r3,r3,r18
		sub			r16,r16,r18
		
		cmpwi		r16,0
		beq			cwidth_fin

start_2_loop:
		lbzu			r12,1(r3)
		li				r13,8
		mtctr		r13
middle_2_loop:
		rlwinm		r18,r12,28,28,28
		srw			r14,r15,r18

		bge			cr7,M2_no_clip

		cmpwi		r10,0
		beq			M2_not_skipping

		subi			r10,r10,1
		addi			r4,r4,1

		b				M2_fin_middle
M2_not_skipping:

		stbu			r14,1(r4)
		subi			r11,r11,1
		cmpwi		r11,0
		beq			next_line

		b				M2_fin_middle
M2_no_clip:
		stbu			r14,1(r4)

M2_fin_middle:
		rlwinm			r12,r12,1,0,30
		
		bdnz			middle_2_loop
		
		subi			r16,r16,1
		cmpwi		r16,0
		bne			start_2_loop

		b			cwidth_fin

do_M_4:
		cmpwi		r16,127
		bne			real_M4
		lbzu			r16,1(r3)
		addi			r16,r16,34 { stupid_compiler_min_42_count_div_4 }
		b				both_M4
		
real_M4:
		subi		r16,r16,93 { stupid_compiler_96_min_4_count_div_4 }
both_M4:
		addi			r3,r3,1
		lswi			r15,r3,4
		addi			r3,r3,3
		
		blt			cr7,are_clipping4

		rlwinm			r18,r16,2,0,29
		sub			r11,r11,r18
		b				start_4_loop

are_clipping4:
		cmpwi		r10,4
		blt			start_4_loop
		
		rlwinm		r18,r16,2,0,29
		rlwinm		r13,r10,0,0,29

		cmpw		r18,r13
		blt			min4
		mr			r18,r13
min4:
		add			r4,r4,r18
		sub			r10,r10,r18
		rlwinm		r18,r18,30,0,31
		add			r3,r3,r18
		sub			r16,r16,r18

		cmpwi		r16,0
		beq			cwidth_fin

start_4_loop:
		lbzu			r12,1(r3)
		li				r13,4
		mtctr		r13
middle_4_loop:
		rlwinm		r14,r12,29,27,28
		slw			r14,r15,r14
		rlwinm		r14,r14,8,24,31

		bge			cr7,M4_no_clip

		cmpwi		r10,0
		beq			M4_not_skipping

		subi			r10,r10,1
		addi			r4,r4,1

		b				M4_fin_middle
M4_not_skipping:

		stbu			r14,1(r4)
		subi			r11,r11,1
		cmpwi			r11,0
		beq			next_line

		b				M4_fin_middle
M4_no_clip:
		stbu			r14,1(r4)

M4_fin_middle:
		rlwinm		r12,r12,2,0,30
		
		bdnz			middle_4_loop
		
		subi			r16,r16,1
		cmpwi		r16,0
		bne			start_4_loop

		b			cwidth_fin

do_RLE:
		cmpwi		r16,251
		bge			RLE_2
		subi			r16,r16,125 { stupid_compiler_128_min_RLE1_count }
		b				RLE_both
RLE_2:
		subi			r16,r16,251
		rlwinm		r16,r16,8,0,23
		lbzu			r18,1(r3)
		or				r16,r16,r18
		addi			r16,r16,126	{ min_RLE2_count }

RLE_both:
		lbzu			r12,1(r3)

		cmpwi		r10,0
		beq			no_RLE_skip
		
		mr			r18,r16
		cmpw		r18,r10
		blt			min5
		mr			r18,r10
min5:
		add			r4,r4,r18
		sub			r16,r16,r18
		sub			r10,r10,r18
		cmpwi		r10,0
		bne			cwidth_fin

no_RLE_skip:
		cmpw		r16,r11
		blt			min6
		mr			r16,r11
min6:
		cmpwi		r16,0
		beq			cwidth_fin
		sub			r11,r11,r16
		mtctr		r16
block_fill_loop:
		stbu			r12,1(r4)
		bdnz			block_fill_loop

cwidth_fin:
//			while cwidth > 0 do begin
		cmpwi		r11,0
		bgt			cwidth_loop

next_line:
		mr			r3,r17
		add			r4,r4,r5
		sub			r4,r4,r7
		sub			r4,r4,r8
		
//		while height > 0 do begin
row_loop_fin:
		subi			r9,r9,1
		cmpwi		r9,0
		bne			row_loop
		
		lwz			r13,-4(sp)
		lwz			r14,-8(sp)
		lwz			r15,-12(sp)
		lwz			r16,-16(sp)
		lwz			r17,-20(sp)
		lwz			r18,-24(sp)
		blr
	end;
{$elsec}
	procedure DoLines( xcur_pos: UInt8Ptr; xaddr: UInt8Ptr; xdest_rowBytes: longint; xorgwidth, xskip, xdowidth, xheight: longint );
	asm;
		const
			param_size = 7*4;
			extra_stack = 8;
			finaddr = 0;
			saved_count = 4;
			stack_size = extra_stack + 64;
			height = stack_size + 0;
			dowidth = stack_size + 4;
			skip = stack_size + 8;
			orgwidth = stack_size + 12;
			dest_rowBytes = stack_size + 16;
			addr = stack_size + 20;
			cur_pos = stack_size + 24;
		const
			stupid_compiler_64_min_2_count_div_8 = 64-min_2_count div 8;
			stupid_compiler_96_min_4_count_div_4 = 96-min_4_count div 4;
			stupid_compiler_min_42_count_div_4 = min_42_count div 4;
			stupid_compiler_128_min_RLE1_count = 128 - min_RLE1_count;
	begin
		movem.l	a0-a6/d0-d7,-(sp)
		subq.l			#extra_stack,sp
{
a5 - addr
a4 - cur_pos

d7 - cskip
d6 - cwidth
d5 - clipping
d4 - b
d3 - jcount
d2 - bt
d1 - b2b1 b1 in bottom half or b4b3b2b1
d0 - count
}
		move.l		cur_pos(sp),a4
		move.l		addr(sp),a5

// 		clipping := dowidth < orgwidth;
		move.l		dowidth(sp),d0
		cmp.l		orgwidth(sp),d0
		slt.b			d5
		
//		while height > 0 do begin
		bra			row_loop_fin
row_loop:

//			BlockMoveData( cur_pos, @data_len, 2 );
//			longint(cur_pos) := longint(cur_pos) + 2;
//			finaddr := UInt8Ptr(longint(cur_pos) + data_len);
		clr.l			d0
		move.b		(a4)+,d0
		lsl.w			#8,d0
		move.b		(a4)+,d0
		move.l		a4,d1
		add.l			d0,d1
		move.l		d1,finaddr(sp)

//			cskip := skip;
		move.l		skip(sp),d7
		
//			cwidth := dowidth;
		move.l		dowidth(sp),d6
		
//			while cwidth > 0 do begin
cwidth_loop:

//				code := cur_pos^;
//				Inc(longint(cur_pos));
		clr.l			d0
		move.b		(a4)+,d0

// case code of
		cmp.b		#64,d0
		blo			do_raw
		cmp.b		#96,d0
		blo			do_M_2
		cmp.b		#128,d0
		blo			do_M_4
		bra			do_RLE

do_raw:
		addq.b		#1,d0
		
		tst.l			d7
		beq			no_raw_skip
		
		move.l		d0,d1
		cmp.l		d1,d7
		bgt			min1
		move.l		d7,d1
min1:
		add.l			d1,a4
		add.l			d1,a5
		sub.l			d1,d0
		sub.l			d1,d7
		bne			cwidth_fin

no_raw_skip:
		cmp.l		d0,d6
		bgt			min2
		move.l		d6,d0
min2:
		tst.l			d0
		beq			cwidth_fin
		sub.l			d0,d6
block_move_loop:
		move.b		(a4)+,(a5)+
		subq.l		#1,d0
		bne			block_move_loop
		
		bra			cwidth_fin	

do_M_2:
		sub.l			#stupid_compiler_64_min_2_count_div_8,d0
		move.b		(a4)+,d1
		swap			d1
		move.b		(a4)+,d1
		swap			d1
		
		tst.b			d5
		bne			are_clipping2
		
		move.l		d0,d3
		lsl.l			#3,d3
		sub.l			d3,d6
		bra			start_2_loop

are_clipping2:
		cmp.l		#8,d7
		blt			start_2_loop
		
		move.l		d0,d3
		lsl.l			#3,d3
		move.l		d7,d4
		and.b			#$F8,d4
		cmp.l		d3,d4
		bgt			min3
		move.l		d4,d3
min3:
		add.l			d3,a5
		sub.l			d3,d7
		lsr.l			#3,d3
		add.l			d3,a4
		sub.l			d3,d0
		beq			cwidth_fin

start_2_loop:
		move.b		(a4)+,d4
		move.b		#8,d3
middle_2_loop:
		move.l		d1,d2
		tst.b			d4
		bge.s			choose_b1
		swap			d2
choose_b1:

		tst.b			d5
		beq.s			M2_no_clip

		tst.l			d7
		beq.s			M2_not_skipping

		subq.l		#1,d7
		addq.l		#1,a5

		bra.s			M2_fin_middle
M2_not_skipping:

		move.b		d2,(a5)+
		subq.l		#1,d6
		beq			next_line

		bra.s			M2_fin_middle
M2_no_clip:
		move.b		d2,(a5)+

M2_fin_middle:
		lsl.b			#1,d4
		
		subq.b			#1,d3
		bne				middle_2_loop
		
		sub.l			#1,d0
		bne			start_2_loop

		bra			cwidth_fin

do_M_4:

		cmp.b		#127,d0
		bne.s			real_M4
		move.b		(a4)+,d0
		add.l			#stupid_compiler_min_42_count_div_4,d0
		bra.s			both_M4
		
real_M4:
		sub.l			#stupid_compiler_96_min_4_count_div_4,d0
both_M4:
		move.b		(a4)+,d1
		ror.l		#8,d1
		move.b		(a4)+,d1
		ror.l		#8,d1
		move.b		(a4)+,d1
		ror.l		#8,d1
		move.b		(a4)+,d1
		ror.l		#8,d1
		
		tst.b			d5
		bne			are_clipping4
		
		move.l		d0,d3
		lsl.l			#2,d3
		sub.l			d3,d6
		bra			start_4_loop

are_clipping4:
		cmp.l		#4,d7
		blt			start_4_loop
		
		move.l		d0,d3
		lsl.l			#2,d3
		move.l		d7,d4
		and.b			#$FC,d4
		cmp.l		d3,d4
		bgt			min4
		move.l		d4,d3
min4:
		add.l			d3,a5
		sub.l			d3,d7
		lsr.l			#2,d3
		add.l			d3,a4
		sub.l			d3,d0
		beq			cwidth_fin

start_4_loop:
		move.l		d0,saved_count(sp)
		move.b		(a4)+,d4
		move.b		#4,d3
middle_4_loop:
		move.b		d4,d0
		lsr.l			#3,d0
		and.l			#$18,d0
		move.l		d1,d2
		lsr.l			d0,d2

		tst.b			d5
		beq.s			M4_no_clip

		tst.l			d7
		beq.s			M4_not_skipping

		subq.l		#1,d7
		addq.l		#1,a5

		bra.s			M4_fin_middle
M4_not_skipping:

		move.b		d2,(a5)+
		subq.l		#1,d6
		beq			next_line

		bra.s			M4_fin_middle
M4_no_clip:
		move.b		d2,(a5)+

M4_fin_middle:
		lsl.b			#2,d4
		
		subq.b			#1,d3
		bne				middle_4_loop
		
		move.l		saved_count(sp),d0
		sub.l			#1,d0
		bne			start_4_loop

		bra			cwidth_fin

do_RLE:
		cmp.b		#251,d0
		bhs.s			RLE_2

		sub.l			#stupid_compiler_128_min_RLE1_count,d0

		bra.s			RLE_both
RLE_2:
		sub.b			#251,d0
		lsl.l			#8,d0
		move.b		(a4)+,d0
		add.l			#min_RLE2_count,d0

RLE_both:
		
		move.b		(a4)+,d4
		
		tst.l			d7
		beq			no_RLE_skip
		
		move.l		d0,d1
		cmp.l		d1,d7
		bgt			min5
		move.l		d7,d1
min5:
		add.l			d1,a5
		sub.l			d1,d0
		sub.l			d1,d7
		bne			cwidth_fin

no_RLE_skip:
		cmp.l		d0,d6
		bgt			min6
		move.l		d6,d0
min6:
		tst.l			d0
		beq			cwidth_fin
		sub.l			d0,d6
		
		cmp.w		#7,d0
		ble.s			slow_fill
		
		move.w		a5,d1
		neg.w		d1
		and.w		#3,d1
		beq			fill_aligned
		sub.w		d1,d0
		subq.w		#1,d1
fill_start:
		move.b		d4,(a5)+
		dbf			d1,fill_start
fill_aligned:
		move.w		d0,d2
		lsr.w		#2,d0 { d0 >= 1 }
		move.b		d4,d1
		lsl.w			#8,d4
		or.b			d1,d4
		move.w		d4,d1
		swap			d4
		move.w		d1,d4
		subq.w		#1,d0
fill_longs:
		move.l		d4,(a5)+
		dbf			d0,fill_longs
		move.w		d2,d0
		and.w		#3,d0
		beq			cwidth_fin
		
slow_fill:
		subq.w		#1,d0
block_fill_loop:
		move.b		d4,(a5)+
		dbf			d0,block_fill_loop

cwidth_fin:
//			while cwidth > 0 do begin
		tst.l			d6
		bgt			cwidth_loop

next_line:
		move.l		finaddr(sp),a4
		
		add.l			dest_rowBytes(sp),a5
		sub.l			skip(sp),a5
		sub.l			dowidth(sp),a5
		
//		while height > 0 do begin
row_loop_fin:
		move.l		height(sp),d0
//			Dec(height);
		subq.l		#1,d0
		move.l		d0,height(sp)
		bge			row_loop
		
		addq.l		#extra_stack,sp
		movem.l	(sp)+,a0-a6/d0-d7
		move.l		(sp),param_size(sp)
		add.l			#param_size,sp
		rts
	end;
{$endc}
{$endc}

	procedure DrawGSIFWithin( data: Handle; dest_pm: PixMapHandle; ath, atv: integer; const bounds: Rect );
		var
			dest_rowBytes: longint;
			dest_addr: Ptr;
			dummy: boolean;
			head: GSIFHeader;
			clip_bounds, pix_bounds: Rect;
			addr: UInt8Ptr;
			line_count: integer;
			height, width, skip: integer;
	begin
		BlockMoveData( data^, @head, SizeOf(head) );
		Assert( head.tag = GSIF_Header_tag );

		dummy := LockPixels( dest_pm );
		dest_addr := dest_pm^^.baseAddr;
		dest_rowBytes := dest_pm^^.rowBytes and $3FFF;
		Assertx( head.width <= dest_pm^^.bounds.right - dest_pm^^.bounds.left );
		Assertx( head.height <= dest_pm^^.bounds.bottom - dest_pm^^.bounds.top );
		
		pix_bounds := dest_pm^^.bounds;
		clip_bounds := bounds;
		if clip_bounds.left < pix_bounds.left then begin
			clip_bounds.left := pix_bounds.left;
		end;
		if clip_bounds.right > pix_bounds.right then begin
			clip_bounds.right := pix_bounds.right;
		end;
		if clip_bounds.top < pix_bounds.top then begin
			clip_bounds.top := pix_bounds.top;
		end;
		if clip_bounds.bottom > pix_bounds.bottom then begin
			clip_bounds.bottom := pix_bounds.bottom;
		end;
		OffsetRect( clip_bounds, -pix_bounds.left, -pix_bounds.top );
		ath := ath - pix_bounds.left;
		atv := atv - pix_bounds.top;
	
		width := Min( clip_bounds.right, ath + head.width ) - Max( ath, clip_bounds.left );
		height := Min( clip_bounds.bottom, atv + head.height ) - Max( atv, clip_bounds.top );
		if (width > 0) & (height > 0) then begin // Something to draw
			addr := UInt8Ptr(longint(data^) + SizeOf(GSIFHeader));
			while atv < clip_bounds.top do begin
				BlockMoveData( addr, @line_count, 2 );
				addr := UInt8Ptr(longint(addr) + line_count + 2);
				Inc(atv);
			end;
			longint(dest_addr) := longint(dest_addr) + atv * dest_rowBytes + ath;
			skip := Max( 0, clip_bounds.left - ath );
			
			DoLines( addr, UInt8Ptr(dest_addr), dest_rowBytes, head.width, skip, width, height );
		end;
	end;
	
	procedure DrawGSIF( data: Handle; dest: PixMapHandle; ath, atv: integer );
	begin
		DrawGSIFWithin( data, dest, ath, atv, dest^^.bounds );
	end;

	procedure GetGSIFSize( gsifdata: Handle; var width, height: integer );
		var
			head: GSIFHeader;
	begin
		Assert( GetHandleSize( gsifdata ) >= SizeOf(GSIFHeader) );
		BlockMoveData( gsifdata^, @head, SizeOf(head) );
		Assert( head.tag = GSIF_Header_tag );
		width := head.width;
		height := head.height;
	end;
	
end.
					if skip = 0 then begin
						count := Min( count, cwidth );
						BlockMoveData( cur_pos, addr, count );
						cwidth := cwidth - count;
						longint(cur_pos) := longint(cur_pos) + count;
						OffsetPtr( addr, ccount );
					end;
						if skip = 0 then begin
							ccount := Min( count, cwidth );
							BlockMoveData( cur_pos, addr, ccount );
							cwidth := cwidth - ccount;
							longint(cur_pos) := longint(cur_pos) + ccount;
							OffsetPtr( addr, ccount );
						end;
					end else begin
						BlockMoveData( cur_pos, addr, count );
						cwidth := cwidth - count;
						longint(cur_pos) := longint(cur_pos) + count;
						OffsetPtr( addr, count );
