unit MyOTThreads;

interface

	uses
		Types, OpenTransport, OpenTptInternet;
	
	const
		kNoTimeout = $FFFFFFFF;
		KHasClosedErr = kOTOutStateErr; // From OTThreadRcv
		
	type
		ThreadConnection = Ptr;

	function ThreadOTConnect( var connection: ThreadConnection; localport: InetPort; const where: Str255; timeout: UInt32 ): OSStatus;
	function ThreadOTSnd( connection: ThreadConnection; data: Ptr; len: UInt32 ): OSStatus;
	function ThreadOTRcv( connection: ThreadConnection; data: Ptr; maxlen: UInt32; var len: UInt32; timeout: UInt32 ): OSStatus;
	function ThreadOTSndOrderlyDisconnect( connection: ThreadConnection ): OSStatus;
	function ThreadOTCloseProvider( var connection: ThreadConnection ): OSStatus;
	
	function ThreadOTGetEndpoint( connection: ThreadConnection; var ep: EndpointRef ): OSStatus;
	function ThreadOTAbort( connection: ThreadConnection ): OSStatus;

implementation

	uses
		Events, TextUtils, OpenTransport, OpenTptInternet, Threads, 
		MyCStrings, MyAssertions, MyOTUtils, MyUtils, MyMemory, MyLowLevel;

	const
		good_magic = $89ABCDEF;
		bad_magic = $DEADBEEF;
	
	type
		ContextRecord = record
			magic:longInt;
			result:OSStatus;
			ep: EndpointRef;
			open_completed: Boolean;
			bind_completed: Boolean;
			connect_received: Boolean;
			disconnect_received: Boolean;
			received_orderly_disconnect: Boolean;
			sent_orderly_disconnect: Boolean;
			abort_wait_on: Boolean;
			rcvcall: TCall;
			rcvsin: InetAddress;
		end;
		ContextPtr = ^ContextRecord;
	
	function ValidContext( context: ContextPtr ): Boolean;
	begin
		ValidContext := ( context <> nil ) & ( context^.magic = good_magic );
	end;
	
	procedure AssertValidContext( context: ContextPtr );
	begin
		Assert( ValidContext( context ) );
	end;
	
	procedure Yield;
		var
			junk: OSErr;
	begin
		junk := YieldToAnyThread;
		Assert( junk = noErr );
	end;
	
	procedure SetupContext( context: ContextPtr );
	begin
		Assert( context <> nil );
		MZero( context, SizeOf(ContextRecord) );
		context^.magic := good_magic;
		AssertValidContext( context );
	end;
	
	function ThreadOTAbort( connection: ThreadConnection ): OSStatus;
		var
			context: ContextPtr;
	begin
		context := ContextPtr(connection);
		if ValidContext( context ) & (context^.ep <> nil) then begin
			context^.abort_wait_on := true;
			ThreadOTAbort := noErr;
		end else begin
			ThreadOTAbort := kOTBadReferenceErr;
		end;
	end;
	
	procedure EventHandler( context: ContextPtr; event: OTEventCode; result: OTResult; cookie: univ Ptr );
		var
//			s: Str15;
			junk: OSStatus;
	begin
		AssertValidContext(context);
		case event of
			T_OPENCOMPLETE: begin
				AddOSStatus( context^.result, result );
				context^.ep := EndpointRef(cookie);
				context^.open_completed := true;
			end;
			T_BINDCOMPLETE:  begin
				AddOSStatus( context^.result, result );
				context^.bind_completed := true;
			end;
			T_CONNECT:  begin
				context^.connect_received := true;
				AddOSStatus( context^.result, result );
				AddOSStatus( context^.result, OTRcvConnect( context^.ep, @context^.rcvCall) );
			end;
			T_DISCONNECT:  begin
				AddOSStatus( context^.result, result );
				context^.disconnect_received := true;
			end;
			T_ORDREL:  begin
				context^.received_orderly_disconnect := true;
				junk := OTRcvOrderlyDisconnect( context^.ep );
			end;
			T_DATA, T_GODATA: begin
			end;
			otherwise begin
				Assert( false );
(*
				s := '';
				if event<0 then begin
					s := '-';
					event := -event;
				end;
				repeat
					s := concat(chr(48 + event mod 10),s);
					event := event div 10;
				until event = 0;
				DebugStr(concat('EventHandler got unexpected event ',s));
*)
			end;
		end;
	end;

	function ThreadOTConnect( var connection: ThreadConnection; localport: InetPort; const where: Str255; timeout: UInt32 ): OSStatus;
		var
			context: ContextPtr;
			err: OSStatus;
			
		procedure WaitOn( var waitfor: boolean );
		begin
			if err = noErr then begin
				while not waitfor and not context^.abort_wait_on do begin
					if (TickCount >= timeout) & (context^.ep <> nil) then begin
						// must not set it during OTAsyncOpenEndpoint
						AddOSStatus( err, kETIMEDOUTErr );
						leave;
					end;
					Yield;
				end;
				AddOSStatus( err, context^.result );
				if context^.abort_wait_on then begin
					AddOSStatus( err, userCanceledErr );
					context^.abort_wait_on := false;
				end;
			end;
		end;
	
		var
			junk: OSStatus;
			sndaddr: DNSAddress;
			sndcall: TCall;
	begin
		if timeout <> kNoTimeout then begin
			timeout := TickCount + timeout;
		end;
		err := MNewPtr( connection, SizeOf( ContextRecord ) );
		if err = noErr then begin
			context := ContextPtr(connection);
			SetupContext( context );

			// OpenEndpoint
			err := OTAsyncOpenEndpoint( MyOTCreateConfiguration( 'tcp' ), 0, nil, @EventHandler, context );
			WaitOn( context^.open_completed );
			
			// Bind
			Assert( localport = 0 );  // NYI
			if err = noErr then begin
				err := OTBind( context^.ep, nil, nil );
				WaitOn( context^.bind_completed );
			end;
			
			// Connect
			if err = noErr then begin
				MZero( @sndcall, SizeOf(sndcall) );
				sndcall.addr.len := MyOTInitDNSAddress( sndaddr, where );
				sndcall.addr.buf := @sndaddr;
				
				MZero( @context^.rcvcall, SizeOf(context^.rcvcall) );
				context^.rcvcall.addr.maxlen := SizeOf(InetAddress);
				context^.rcvcall.addr.buf := @context^.rcvsin;
	
				err := OTConnect( context^.ep, @sndcall, @context^.rcvcall );
				if err = kOTNoDataErr then begin
					err := noErr;
				end;
				WaitOn( context^.connect_received );
			end;
			
			if err = noErr then begin
				AddOSStatus( err, OTSetSynchronous( context^.ep ) );
				AddOSStatus( err, OTSetNonBlocking( context^.ep ) );
			end;
		
			if err <> noErr then begin
				if context^.ep <> nil then begin
					junk := OTCloseProvider( context^.ep );
					context^.ep := nil;
					context^.magic := bad_magic;
				end;
				MDisposePtr( connection );
			end;
		end;
		ThreadOTConnect := err;
	end;

	function ThreadOTSnd( connection: ThreadConnection; data: Ptr; len: UInt32 ): OSStatus;
		var
			context: ContextPtr;
			result: OTResult;
			again:Boolean;
	begin
		context := ContextPtr(connection);
		if ValidContext( context ) then begin
			result := noErr;
			if len > 0 then begin
				repeat
					result := OTSnd( context^.ep, data, len, 0 );
					if result > 0 then begin
						OffsetPtr( data, result );
						len := len - result;
						result := noErr;
					end;
					again := ((result = kOTFlowErr) or (result = noErr)) and (len > 0);
					if again then begin
						Yield;
					end;
				until context^.abort_wait_on or not again;
			end;
			if context^.abort_wait_on then begin
				AddOSStatus( result, userCanceledErr );
				context^.abort_wait_on := false;
			end;
		end else begin
			result := kOTBadReferenceErr;
		end;
		ThreadOTSnd := result;
	end;

	function ThreadOTRcv( connection: ThreadConnection; data: Ptr; maxlen: UInt32; var len: UInt32; timeout: UInt32 ): OSStatus;
		var
			context: ContextPtr;
			result: OTResult;
			again:Boolean;
			flags: OTFlags;
	begin
		context := ContextPtr(connection);
		if ValidContext( context ) then begin
			if timeout <> kNoTimeout then begin
				timeout := TickCount + timeout;
			end;
			result := noErr;
			len := 0;
			while (maxlen > 0) & not context^.abort_wait_on do begin
				result := OTRcv( context^.ep, data, maxlen, flags );
				if result = kOTNoDataErr then begin
					result := 0;
				end;
				if result > 0 then begin
					OffsetPtr( data, result );
					len := len + result;
					maxlen := maxlen - result;
					result := noErr;
				end;
				again := (result = noErr) and (maxlen > 0) and (TickCount < timeout);
				if again then begin
					Yield;
				end else begin
					leave;
				end;
			end;
			if context^.abort_wait_on then begin
				AddOSStatus( result, userCanceledErr );
				context^.abort_wait_on := false;
			end;
		end else begin
			result := kOTBadReferenceErr;
		end;
		ThreadOTRcv := result;
	end;
	
	function ThreadOTSndOrderlyDisconnect( connection: ThreadConnection ): OSStatus;
		var
			context: ContextPtr;
			err: OSStatus;
	begin
		context := ContextPtr(connection);
		if ValidContext( context ) then begin
			err := OTSndOrderlyDisconnect( context^.ep );
		end else begin
			err := kOTBadReferenceErr;
		end;
		ThreadOTSndOrderlyDisconnect := err;
	end;
	
	function ThreadOTCloseProvider( var connection: ThreadConnection ): OSStatus;
		var
			context: ContextPtr;
			err: OSStatus;
	begin
		context := ContextPtr(connection);
		if ValidContext( context ) then begin
			if context^.ep <> nil then begin
				err := OTCloseProvider( context^.ep );
				context^.ep := nil;
			end else begin
				err := noErr;
			end;
			context^.magic := bad_magic;
			MDisposePtr( connection );
		end else begin
			err := kOTBadReferenceErr;
		end;
		connection := nil;
		ThreadOTCloseProvider := err;
	end;
	
	function ThreadOTGetEndpoint( connection: ThreadConnection; var ep: EndpointRef ): OSStatus;
		var
			context: ContextPtr;
			err: OSStatus;
	begin
		context := ContextPtr(connection);
		if ValidContext( context ) & (context^.ep <> nil) then begin
			ep := context^.ep;
			err := noErr;
		end else begin
			err := kOTBadReferenceErr;
		end;
		ThreadOTGetEndpoint := err;
	end;
	
end.
	
	function ThreadOTOpenEndpoint(config:Str255; notifier: OTNotifyProcPtr; context:ContextPtr):OSStatus;
		var
			err:OSStatus;
			info: TEndpointInfo;
	begin
		SetupContext(context);
		P2C(@config);
		context^.open_completed := false;
		err := OTAsyncOpenEndpoint(OTCreateConfiguration(@config), 0, info, notifier, context);
		if err = noErr then begin
			while not context^.open_completed do begin
				MyYield;
			end;
			err := context^.result;
		end;
		ThreadOTOpenEndpoint := err;
	end;

	function ThreadOTBind(context:ContextPtr; reqAddr: TBindPtr; retAddr: TBindPtr):OSStatus;
		var
			err:OSStatus;
	begin
		context^.bind_completed := false;
		err := OTBind(context^.ep, reqAddr, retAddr);
		if err = noErr then begin
			while not threads_must_die & not context^.bind_completed do begin
				MyYield;
			end;
			if context^.bind_completed then begin
				err := context^.result;
			end else begin
				err := userCanceledErr;
			end;
		end;
		ThreadOTBind := err;
	end;
	 
	function ThreadOTConnect(context:ContextPtr; var sndCall: TCall; var rcvCall: TCall): OSStatus;
		var
			err:OSStatus;
			junk:OSStatus;
	begin
		context^.connect_received := false;
		context^.disconnect_received := false;
		context^.sent_orderly_disconnect := false;
		context^.received_orderly_disconnect := false;
		err := OTConnect(context^.ep, sndCall, rcvCall);
		if err = kOTNoDataErr then begin
			err := noErr;
		end;
		if err = noErr then begin
			while not threads_must_die & not context^.disconnect_received & not context^.connect_received do begin
				MyYield;
			end;
			if context^.disconnect_received then begin
				junk := OTRcvDisconnect(context^.ep, nil);
				err := -1;
			end else if context^.connect_received then begin
				err := context^.result;
				if err = noErr then begin
					err := OTRcvConnect(context^.ep,rcvCall);
				end;
			end else begin
				err := userCanceledErr;
			end;
		end;
		ThreadOTConnect := err;
	end;

	function ThreadOTSnd(context:ContextPtr; buf: univ Ptr; nbytes: size_t): OTResult;
		var
			result: OTResult;
			again:Boolean;
	begin
		result := noErr;
		if nbytes > 0 then begin
			repeat
				result := OTSnd(context^.ep,buf,nbytes,0);
				if result > 0 then begin
					longint(buf) := longint(buf) + result;
					nbytes := nbytes - result;
					result := noErr;
				end;
				again := ((result = kOTFlowErr) or (result = noErr)) and (nbytes > 0);
				if again then begin
					MyYield;
				end;
			until threads_must_die or not again;
		end;
		ThreadOTSnd := result;
	end;

	function ThreadOTSndOrderlyDisconnect(context:ContextPtr):OSStatus;
		var
			err:OSErr;
	begin
		err := noErr;
		if not context^.sent_orderly_disconnect then begin
			err := OTSndOrderlyDisconnect(context^.ep);
			context^.sent_orderly_disconnect := true;
		end;
		ThreadOTSndOrderlyDisconnect := err;
	end;

	function ThreadOTCloseProvider(context:ContextPtr; clean:Boolean):OSStatus;
		var
			err:OSStatus;
			result:OTResult;
			timeout:longInt;
	begin
		if clean then begin
			err := ThreadOTSndOrderlyDisconnect(context);
			if err = noErr then begin
				timeout := TickCount + 3*60*60;
				repeat
					result:=OTGetEndpointState(context^.ep);
				until (result < 0) or (result = T_UNINIT) or (result = T_IDLE) or (TickCount > timeout);
			end;
		end;
		err := OTCloseProvider(context^.ep);
		ThreadOTCloseProvider := err;
	end;
	
	procedure TestFinger(ignored:Ptr);
		var
			err, err2: OSStatus;
			mystr, remotename, space : Str255;
			sndaddr: DNSAddress;
			sndcall, rcvcall: TCall;
			rcvsin: InetAddress;
			flags:OTFlags;
			bytes: longInt;
			contextspace:ContextRecord;
			context:ContextPtr;
	begin
		context := @contextspace;

		bytes := 0;
		flags := 0;

		remotename := 'guppy:79';
		
		mystr := kTCPName;
		err := ThreadOTOpenEndpoint(kTCPName, @EventHandler, context);
		if (err <> noErr) then begin
			writeln('ERROR: ThreadOTOpenEndpoint failed with ', err : 1);
		end;

		if err = noErr then begin
			err := ThreadOTBind(context, nil, nil);
			if (err <> noErr) then begin
				writeln('ERROR: TheadOTBind failed with ', err : 1);
			end;
	
			if err = noErr then begin
				ClearBlock(@sndcall, sizeof(sndcall));
				P2C(@remotename);
				sndcall.addr.len := OTInitDNSAddress(sndaddr, @remotename);
				sndcall.addr.buf := @sndaddr;
				
				ClearBlock(@rcvcall, sizeof(rcvcall));
				rcvcall.addr.maxlen := sizeof(InetAddress);
				rcvcall.addr.buf := @rcvsin;
	
				err := ThreadOTConnect(context, sndcall, rcvcall);
				if (err <> noErr) then begin
					writeln('ERROR: Connect() failed with ', err : 1);
				end;
			end;
	
			if err = noErr then begin
				space := concat('peter',chr(13),chr(10));
				err := ThreadOTSnd(context, @space[1], length(space));
				if (err <> noErr) then begin
					writeln('ERROR: Snd() failed with ', err : 1);
				end;
			end;
	
			if err = noErr then begin
				repeat
					MyYield;
					bytes := OTRcv(context^.ep, @space[1], 255, flags);
					if bytes > 0 then begin
						space[0] := chr(bytes);
						write(space);
					end else if bytes <> kOTNoDataErr then begin
						writeln('OTRcv ',bytes);
					end;
				until ((bytes<0) & (bytes <> kOTNoDataErr)) | context^.received_orderly_disconnect | threads_must_die;
				if context^.received_orderly_disconnect then begin
					bytes := noErr;
				end;
				if bytes < 0 then begin
					SetContextError(context, bytes);
				end;
				err := context^.result;
			end;

			err2 := ThreadOTCloseProvider(context, err = noErr);
			if (err2 <> noErr) then begin
				writeln('ERROR: ThreadOTCloseProvider failed with ', err2 : 1);
			end;
			if (err = noErr) then begin
				err := err2;
			end;
		end;

		writeln('Bye');
		quitNow :=true;
		MyThreadDied;
	end;

	procedure TestFingerX(ignored:Ptr);
		var
			i:integer;
	begin
		for i := 1 to 3 do begin
			while not button do begin
				MyYield
			end;
			while button do begin
				MyYield
			end;
		end;
		quitNow :=true;
		MyThreadDied;
	end;
	
	var
		err:OSStatus;
begin
	Initialization;
	writeln;
	err := InitOpenTransport;
	writeln('InitOpenTransport ',err:1);
	if err=noErr then begin
		quitNow := false;
		InitMyThreads;
		err := MyNewThread (@TestFinger, nil); { increments active_thread }

		if err = noErr then begin
			while not quitNow do begin
				WNE;
				MyYield;
			end;
		end;
		
		FinishMyThreads;
		CloseOpenTransport;
	end;
end.
